/*
 * FILE INFO: s3c2440fpga.c 
 *                  FPGA
 *                  ںlinux-2.6.13
 * Date        : 20081104
 * ---------------------------
 * modify : 
 * date:     20090111 
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/arch/map.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/page.h>
#include <linux/mm.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/regs-dma.h>
#include <asm/arch/regs-irq.h>

#define BEGIN_KMEM {mm_segment_t old_fs = get_fs();set_fs(KERNEL_DS);
#define END_KMEM 	set_fs(old_fs);}

#define DEBUG 1
#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(fmt, ##args)
#else
#define DPRINTK(fmt, args...)
#endif

#define clear_pending(bit) { \
	writel(bit, S3C2410_SRCPND);\
	writel(bit, S3C2410_INTPND);\
    readl(S3C2410_INTPND); \
}


#define clear_pending(bit) {}

#define FIFO_DMA_MAX_LEN	(1024*1024)
#define INT_EINT3_NUM	(1<<3)	
#define INT_DMA0_NUM    (1<<17)
#define DEVICE_NAME		"dso-fpga"
#define DEVICE_MAJOR	(245)
#define DEVICE_MINOR	(0)

#define DMA_DONE	(1)
#define DMA_NODONE	(0)

#define FPGA_BANK_PHY_ADDR	(0x10000000)
#define DEMAND	0
#define HANDSHAKE 1

#define SINGLE 0
#define WHOLE 1

#define BYTE 0
#define HALF_WORD 1
#define ONE_WORD 2

#define UNIT 0
#define BURST 1
#define MEM_LEN 4000
#define DEV_IOC_MAGIC	'd'
#define DEV_IOC_MAXNR	(15)

#define DEV_IOC_TURNON_DMA		_IO(DEV_IOC_MAGIC,  0)
#define DEV_IOC_TURNOFF_DMA		_IO(DEV_IOC_MAGIC,  1)
#define DEV_IOC_SET_FIFO		_IO(DEV_IOC_MAGIC,  2)     //fifo Ч
#define DEV_IOC_CLR_DMA_FLAG	_IO(DEV_IOC_MAGIC,  3)     //DMA ־
#define DEV_IOC_ENABLE_DMA		_IO(DEV_IOC_MAGIC,  4)


static int *mem_pt;
static unsigned int *fpga_data_pt;
static unsigned int *test_data_pt;

static unsigned dma_addr;
static unsigned test_addr;

static int data_need_size = 0;
static int dma_done_flag = DMA_NODONE;
static DECLARE_WAIT_QUEUE_HEAD(wq);
static int read_flag = 0;
static char *src_addr; 	// image addr
static char *dest_addr;		// image addr
static int phy_src_addr;
static int phy_dest_addr;


#define DRIVER_LOG_FILE "/fpgabank.conf"
#define FPGA_221422 "22142"
#define FPGA_22822  "22822"
#define FPGA_11811  "11811"
#define FPGA_01800  "01800"

static int read_bankconftext(void)
{
	 unsigned char  buf[50];
	 int fd;
	 int ret = 0;
	struct file *logo_fd;
	
	
	//logo_fd = filp_open(logo_1, O_RDONLY | O_LARGEFILE, 0);
	
	 memset(buf, 0, sizeof(buf));
	 //memcpy(buf, DRIVER_VERSION, sizeof(DRIVER_VERSION));
BEGIN_KMEM
	 fd = sys_open(DRIVER_LOG_FILE, O_RDWR, 0);
	 if(fd>=0)
	 {
		  sys_read(fd, buf, 10);
		  //printk("buf : %s\n", buf);
		  if(strstr(buf, FPGA_221422)) 			ret = 1;
		  else if (strstr(buf, FPGA_22822))		ret = 2;
		  else if (strstr(buf, FPGA_11811))		ret = 3;
		  else if (strstr(buf, FPGA_01800))     ret = 4;
		  sys_close(fd);
	 }
	 else
	 	printk("open driver log file error!!\n");
END_KMEM

	return ret;
}



static void init_fpga_bank(void)
{
	//bank2 16λ
	writel(readl(S3C2410_BWSCON)&(~(0xf00)), S3C2410_BWSCON);
	
	writel(readl(S3C2410_BWSCON)|(0x100), S3C2410_BWSCON);
	DPRINTK("bwscon:0x%x\n", readl(S3C2410_BWSCON));
	
	
	//дʱ22622
	switch(read_bankconftext())
	{
		case 1:
			writel(S3C2410_BANKCON_PMCnorm|S3C2410_BANKCON_Tcah2|S3C2410_BANKCON_Tcoh2|S3C2410_BANKCON_Tacc14|
		   		S3C2410_BANKCON_Tcos2|S3C2410_BANKCON_Tacs2, S3C2410_BANKCON2);
			printk("fpga bank 221422\n");
			break;
		case 2:
			writel(S3C2410_BANKCON_PMCnorm|S3C2410_BANKCON_Tcah2|S3C2410_BANKCON_Tcoh2|S3C2410_BANKCON_Tacc8|
		   		S3C2410_BANKCON_Tcos2|S3C2410_BANKCON_Tacs2, S3C2410_BANKCON2);
			printk("fpga bank 22822\n");
			break;
		case 3:
			writel(S3C2410_BANKCON_PMCnorm|S3C2410_BANKCON_Tcah1|S3C2410_BANKCON_Tcoh1|S3C2410_BANKCON_Tacc8|
		   		S3C2410_BANKCON_Tcos1|S3C2410_BANKCON_Tacs1, S3C2410_BANKCON2);
			printk("fpga bank 11811\n");
			break;
		case 4:
			writel(S3C2410_BANKCON_PMCnorm|S3C2410_BANKCON_Tcah0|S3C2410_BANKCON_Tcoh1|S3C2410_BANKCON_Tacc8|
		   		S3C2410_BANKCON_Tcos0|S3C2410_BANKCON_Tacs0, S3C2410_BANKCON2);
			printk("fpga bank 01800\n");
			break;
		default:
			writel(S3C2410_BANKCON_PMCnorm|S3C2410_BANKCON_Tcah2|S3C2410_BANKCON_Tcoh2|S3C2410_BANKCON_Tacc14|
		   		S3C2410_BANKCON_Tcos2|S3C2410_BANKCON_Tacs2, S3C2410_BANKCON2);
			printk("fpga bank 221422\n");
			break;
	}

	writel(readl(S3C2410_GPACON)|0x2000, S3C2410_GPACON);
	writel(readl(S3C2410_GPADAT)|(1<<13), S3C2410_GPADAT);
	
}

/*
 * mmap bank2ûռ䣬2.4ںʹremap_page_range
 * 2.6ںremap_pfn_range
 * עremap_pfn_rangeΪַpage_shiftȵ
*/
static int mmap_fpga_bank2(struct file *file, struct vm_area_struct *vma)
{
	
	//vma->vm_pgoff = virt_to_phys(mem_pt)>>PAGE_SHIFT;
	vma->vm_pgoff = FPGA_BANK_PHY_ADDR>>PAGE_SHIFT;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	
	vma->vm_flags |= VM_RESERVED;

	if(remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff, vma->vm_end-vma->vm_start, vma->vm_page_prot))
	{
		return -EAGAIN;
	}
	

	return 0;
}


/*
 *  GPF3  EINT3ΪFIFOжϱ־ź
 *   EINT3Ϊжź
 */
static void init_gpio()
{
	writel(readl(S3C2410_GPFCON)&(0xffffff3f), S3C2410_GPFCON);
	writel(readl(S3C2410_GPFCON)|(0x80), S3C2410_GPFCON);
	writel(readl(S3C2410_EXTINT0)&(0xffff0fff), S3C2410_EXTINT0);
	writel(readl(S3C2410_EXTINT0)|(0x4000), S3C2410_EXTINT0);	  

	
}

void do_dma(int tc, int dsz, int burst, int servmode, int dmd_hs)
{

#if 0
	writel(test_addr, DISRC0);
	writel((0<<1)|(0<<0), DISRCC0);
	writel(dma_addr, DIDST0);
	writel((0<<1)|(0<<0), DIDSTC0);
	//DMAжģʽ bit 291
	writel((dmd_hs<<31)|(1<<30)|(0<<29)|(burst<<28)|(servmode<<27)|(0<<23)|(1<<22)|(dsz<<20)|(tc), DCON0 ); 
		writel((1<<1)|1, DMTRIG0);
		DPRINTK("now start dma0 transer\n");
	while(0 != readl(DSTAT0));
#endif

writel(FPGA_BANK_PHY_ADDR+(0x09<<2), DISRC0);
writel((0<<1)|(1<<0), DISRCC0);
writel(dma_addr, DIDST0);
writel((0<<1)|(0<<0), DIDSTC0);
writel((1<<31)|(1<<30)|(1<<29)|(burst<<28)|(1<<27)|(0<<23)|(1<<22)|(dsz<<20)|(tc), DCON0);
writel((1<<1)|1, DMTRIG0);
//while(0 != readl(DSTAT0));


}


/*
 * irq server routine
 */

void fifo_full_irq(int irq, void *dev_id, struct pt_regs *regs)
{
   clear_pending(INT_EINT3_NUM);
   do_dma(data_need_size/16, ONE_WORD, BURST, SINGLE, HANDSHAKE);
  
   DPRINTK("get fifo full irq\n");

}


void dma0_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;
   clear_pending(INT_DMA0_NUM);
   DPRINTK("dma0_irq: dma0 transer over\n");

 for (i=0;i<10;i++)
	DPRINTK("0x%x, ", *(fpga_data_pt+i));
   dma_done_flag = DMA_DONE;
  // wake_up_interruptible(&wq);

}


/*
 *  readã һread ʱȡDMAϱ־λ
 *                                           ڶreadʱȡDMAڴ
 */
static ssize_t read_fpga_data(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
	DPRINTK("READ\n");
	if (0 == read_flag)
	{
		wait_event_interruptible(wq, dma_done_flag==DMA_DONE);
		DPRINTK("after wait\n");
		copy_to_user(buffer, (const void *)&dma_done_flag, sizeof(dma_done_flag));
		dma_done_flag = DMA_NODONE;
		read_flag = 1;
	}
	else
	{
		copy_to_user(buffer, (const void *)fpga_data_pt, data_need_size);
		read_flag = 0;
	}
		
	return 0;

}



static int set_irq(void)
{
	int retvalue;
	
	if (request_irq(IRQ_EINT3, fifo_full_irq, SA_INTERRUPT, "fpga fifo full", NULL))
	{
		printk("cannot get irq ent3 for fpga \n");
		return -1;
	}
	
	if (request_irq(IRQ_DMA0, dma0_irq, SA_INTERRUPT, "dma0 transfer interrupt", NULL))
	{
		printk("cannot get irq for dma0 \n");
		return -1;
	}
	
	//writel(readl(S3C2410_MISCCR)&(0xffffff3f), S3C2410_MISCCR);
	writel(readl(S3C2410_INTMSK) &(~(INT_EINT3_NUM)), S3C2410_INTMSK);
	writel(readl(S3C2410_INTMSK) &(~(INT_DMA0_NUM)), S3C2410_INTMSK);
	printk("get irq ent3, irq dma0 for fpga ok\n");
	clear_pending(INT_EINT3_NUM);
	return 0;
}



static int dma_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret;
	
	if (_IOC_TYPE(cmd) != DEV_IOC_MAGIC)
	{
		return -ENOTTY;
	}
	if (_IOC_NR(cmd) > DEV_IOC_MAXNR) 
	{
		return -ENOTTY;
	}

	switch(cmd)
	{
		case DEV_IOC_TURNON_DMA:
		//	writel((1<<1)|1, DMTRIG0);
		//	while(0 != readl(DSTAT0));
			ret = 0;
		break;
		case DEV_IOC_SET_FIFO:     //FIFO ]
		    
			if (arg <0)
				return -1;
			data_need_size = arg;
			DPRINTK("set fifo len:0x%x\n", data_need_size);
			
		break;
		case DEV_IOC_CLR_DMA_FLAG:
			dma_done_flag = DMA_NODONE;
		break;
		case DEV_IOC_ENABLE_DMA:
			 DPRINTK("enable dma transfer\n");
			 do_dma(4000/16, ONE_WORD, BURST, SINGLE, HANDSHAKE);
		break;
		default:
			ret = -1;
		break;
	}

	return ret;
}


static struct file_operations fpga_fops =
{
	owner:		THIS_MODULE,
	mmap:		mmap_fpga_bank2,
	ioctl:		dma_ioctl,
	read:		read_fpga_data,

};


static int __init fpga_init(void)
{
	int ret;
/*
	fpga_data_pt = (unsigned int *)dma_alloc_coherent(NULL, FIFO_DMA_MAX_LEN, &dma_addr, GFP_KERNEL);
	test_data_pt = (unsigned char *)dma_alloc_coherent(NULL, PAGE_SIZE, &test_addr, GFP_KERNEL);
	memset(test_data_pt, 0x52, PAGE_SIZE);
	if (NULL == fpga_data_pt)
	{
		printk("alloc dma mem error\n");
		return -ENOMEM;
	}
*/
	init_fpga_bank();
	init_gpio();
	//set_irq();

	//src_addr = kmalloc(MEM_LEN*sizeof(char), GFP_KERNEL ^ GFP_DMA);    
	//dest_addr = kmalloc(MEM_LEN*sizeof(char), GFP_KERNEL ^ GFP_DMA);
	//phy_src_addr = virt_to_phys(src_addr);

	//phy_dest_addr = virt_to_phys(dest_addr);
	//memset(src_addr, 0x23, MEM_LEN);
	
	ret = register_chrdev(DEVICE_MAJOR, DEVICE_NAME, &fpga_fops);
	
	if (ret <0)
	{
		printk(DEVICE_NAME ": can't register device\n");
		return ret;
	}
	devfs_mk_cdev(MKDEV(DEVICE_MAJOR, DEVICE_MINOR), S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP, DEVICE_NAME);
	
	printk(DEVICE_NAME ": install ok\n");
	return 0;
}

static void __exit fpga_exit(void)
{
	dma_free_coherent(NULL, FIFO_DMA_MAX_LEN, fpga_data_pt, dma_addr);
	devfs_remove(DEVICE_NAME);
	unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
	free_irq(IRQ_EINT3, NULL);
	free_irq(IRQ_DMA0, NULL);
	printk(DEVICE_NAME ": uninstall\n");
}


module_init( fpga_init );
module_exit( fpga_exit );
MODULE_LICENSE("GPL");    
MODULE_AUTHOR("TEKWAY");

 
