/*
 * FILE INFO: s3c2440-i2c.c 
 *                  I2C
 *                  ںlinux-2.6.13
 *                  I2CAT24C08
 * Date        : 20081112
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/arch/map.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/page.h>
#include <linux/mm.h>
#include <asm/arch/regs-iic.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-irq.h>
#include <linux/device.h>


#define DRIVER_VERSION  "[I2C] 1.00.0(090713)\n"
#define DRIVER_LOG_FILE "/dso/driver/driver.log"
#define BEGIN_KMEM {mm_segment_t old_fs = get_fs();set_fs(KERNEL_DS);
#define END_KMEM 	set_fs(old_fs);} 

#define DEBUG 1
#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(fmt, ##args)
#else
#define DPRINTK(fmt, args...)
#endif

#define DEV_IOC_MAGIC	'i'
#define DEV_IOC_MAXNR	(15)

#define DEV_IOC_SET_I2C_BANDRATE		_IO(DEV_IOC_MAGIC,  0)
#define DEV_IOC_SET_I2C_SLAVE_ADDR		_IO(DEV_IOC_MAGIC,  1)


#define DEVICE_NAME		"dso-i2c"
#define DEVICE_MAJOR	(249)
#define DEVICE_MINOR	(0)

#define I2C_CLK			S3C2410_GPE14
#define I2C_DAT			S3C2410_GPE15

#define INT_IIC					  (1 << 27)	
#define I2C_REGS_BASE_PHYS_ADDR   (0x54000000)
#define I2C_REGS_CNT			  (5)


#define rIICDS  (i2c->regs + S3C2410_IICDS)
#define rIICSTAT (i2c->regs + S3C2410_IICSTAT)
#define rIICCON (i2c->regs + S3C2410_IICCON)


#define clear_pending(bit) { \
	writel(bit, S3C2410_SRCPND);\
	writel(bit, S3C2410_INTPND);\
    readl(S3C2410_INTPND); \
}


enum s3c2440_i2c_state
{
	I2C_STATE_IDLE = 0,
	I2C_STATE_WRDATA,
	I2C_STATE_POLLACK,
	I2C_STATE_RDDATA,
	I2C_STATE_SETADDR,
};

typedef struct {
	struct clk			*clk;
	struct device 		*dev;
	void __iomem		*regs;
	struct resource		*irq;
	struct resource		*ioarea;
}i2c_dev_t;

i2c_dev_t local_i2c_dev;
i2c_dev_t *i2c = &local_i2c_dev;
static unsigned char slave_i2c_addr = 0xa0;
#define I2C_PAGE_SIZE		(8)
typedef struct
{
	unsigned char opt_addr;
	unsigned char data_buf[PAGE_SIZE + 4];
	unsigned int data_count;
	unsigned int data_pt;
	unsigned int state;
	unsigned int recv_ack;
}i2c_opt_t;

i2c_opt_t local_i2c_opt;
i2c_opt_t *i2c_opt = &local_i2c_opt;

static int init_s3c2440_i2c(void)
{

	//s3c2410_gpio_cfgpin(I2C_DAT, S3C2410_GPE15_IICSDA);
	//s3c2410_gpio_cfgpin(I2C_CLK, S3C2410_GPE14_IICSCL);
	writel((readl(S3C2410_GPECON)&0x0fffffff)|0xa0000000, S3C2410_GPECON);
	writel(0x0000, S3C2410_GPEUP);
	DPRINTK("S3C2410_GPECON : 0X%x , S3C2410_GPEUP: 0X%x\n", readl(S3C2410_GPECON), readl(S3C2410_GPEUP));
	writeb(0xaf, i2c->regs + S3C2410_IICCON);
	writeb(0x10, i2c->regs + S3C2410_IICADD);
	writeb(0x10, i2c->regs + S3C2410_IICSTAT);
	
	writel(readl(S3C2410_INTMSK) &(~(INT_IIC)), S3C2410_INTMSK);
	clear_pending(INT_IIC);
	return 0;
	
}


static irqreturn_t  i2c_int_hndlr(int irq, void *dev_id, struct pt_regs *regs)
{

		int i;
		clear_pending(INT_IIC);
	
		switch(i2c_opt->state)
		{
		   case I2C_STATE_POLLACK:				 	 //ȴACKģʽ
				i2c_opt->recv_ack = readb(rIICSTAT);  //IICSTAT0λʾǷյACK
			break;
	
		   case I2C_STATE_RDDATA: 						   //ģʽ
			   if((i2c_opt->data_count--)==0) 		   //ֻҪȡ1ֽ
			   {
				   i2c_opt->data_buf[i2c_opt->data_pt++] =  readb(rIICDS);
					writeb(0x90, rIICSTAT);			    //ֹͣ
					writeb(0xe0, rIICCON);
				   	udelay(10);					    
				  	break;	 
			   }	  
			   i2c_opt->data_buf[i2c_opt->data_pt++] = readb(rIICDS); 		//The last data has to be read with no ack.
	
			   if((i2c_opt->data_count)==0)			 
				   writeb(0x60, rIICCON);				
			   else 
				  writeb(0xe0, rIICCON);
				   break;
	
			case I2C_STATE_WRDATA:							//дģʽ
				if((i2c_opt->data_count--)==0)			//д
				{
					writeb(0xd0, rIICSTAT);
					writeb(0xe0, rIICCON);
					udelay(100);						
					break;	  
				}
		
				writeb(i2c_opt->data_buf[i2c_opt->data_pt++], rIICDS);
				for(i=0;i<10;i++);					
				  writeb(0xe0, rIICCON);
				
				break;
	
			case I2C_STATE_SETADDR: 						 // öַģʽ
				if((i2c_opt->data_count--)==0)
					break;						
					writeb(i2c_opt->data_buf[i2c_opt->data_pt++], rIICDS);
				for(i=0;i<10;i++);					
				writeb(0xe0, rIICCON);
				break;
	
			default:
				break;		
		}

	return IRQ_HANDLED;



	
}




static void i2c_write_byte(unsigned int addr, unsigned char data)
{
	i2c_opt->state = I2C_STATE_WRDATA;
	i2c_opt->data_pt = 0;
	i2c_opt->data_buf[1] = (unsigned char)addr;
	i2c_opt->data_buf[0] = (unsigned char)(addr>>8);
	i2c_opt->data_buf[2] = data;
	i2c_opt->data_count = 3;

	//DPRINTK("i2c_opt->data_buf[0] : %d %d \n", i2c_opt->data_buf[0] ,i2c_opt->data_buf[1]);
	writeb(slave_i2c_addr, rIICDS);
	writeb(0xf0, rIICSTAT);
 	while(i2c_opt->data_count != -1);
   	i2c_opt->state = I2C_STATE_POLLACK;
	
	while(1)
    {
		writeb(slave_i2c_addr, rIICDS);
  		i2c_opt->recv_ack = 0x100;
	 	writeb(0xf0, rIICSTAT);    //MasTx,Start
      	writeb(0xaf, rIICCON);
    	while(i2c_opt->recv_ack == 0x100);
        if(!(i2c_opt->recv_ack & 0x1))
           break;                      //ACK
    }

	writeb(0xd0, rIICSTAT);		//ֹͣTX	
	writeb(0xaf, rIICCON);
    udelay(10);                      

}



static void i2c_read_byte(unsigned int addr, unsigned char *data)
{

	 i2c_opt->state = I2C_STATE_SETADDR;
	 i2c_opt->data_pt = 0;
	 i2c_opt->data_buf[1] = (unsigned char)addr;
	 i2c_opt->data_buf[0] = (unsigned char)(addr>>8);
	 i2c_opt->data_count = 2;
	
	//DPRINTK("[K ] I2C READ ADDR :%d\n", addr);
	//DPRINTK("i2c_opt->data_buf[0] : %d %d \n", i2c_opt->data_buf[0] ,i2c_opt->data_buf[1]);
	 writeb(slave_i2c_addr, rIICDS);

	 writeb(0xf0, rIICSTAT);	
     while(i2c_opt->data_count != -1);

	 i2c_opt->state 	= I2C_STATE_RDDATA;
	 i2c_opt->data_pt		 = 0;
	 i2c_opt->data_count  = 1;
	 writeb(slave_i2c_addr, rIICDS);
	 writeb(0xb0, rIICSTAT);
	 writeb(0xaf, rIICCON);			   //Resumes IIC operation.   
	 while(i2c_opt->data_count!=-1);

	 *data = i2c_opt->data_buf[1];
	// DPRINTK("data:0x%x\n", *data);


}


#define OPT_DATA_MAX_SIZE	(256 + 2)
static unsigned char data_buffer[PAGE_SIZE + 2];
static unsigned int access_addr = 0;
static unsigned int g_addr = 0;
static ssize_t s3c2440_i2c_write(struct file *file, char __user *buf, size_t count, loff_t *offset)
{
	int i;

	g_addr = (buf[0]|buf[1]<<8);

	if ((count < 0) || (count > OPT_DATA_MAX_SIZE))
	{
		printk("too large size data\n");
		return -1;
	}

	if(count == 3)
	{
		i2c_write_byte(g_addr, buf[2]);
	}

	#if 0
	for (i=2; i<(count); i++)
	{
		//DPRINTK(" write addr : %d\n", g_addr);
		i2c_write_byte(g_addr++, buf[i]);
		//	DPRINTK(" buf[i]: 0x%x\n", buf[i]);

	}
	#endif

	return count;
}

static ssize_t s3c2440_i2c_read(struct file *file, char __user *buf, size_t count, loff_t *offset)
{
	int i;

	if ((count < 0) || (count > OPT_DATA_MAX_SIZE))
	{
		printk("too large size data\n");
		return -1;
	}
	

	i2c_read_byte(g_addr, buf);

	return count;
	
}

/*
 * -------------------------------------------------------------------------
 *  Function:    s3c2440_i2c_probe   
 * ------------------------------------------------------------------------- 
 *  Input :  
 *  Output: 
 *  Return: 
 * ------------------------------------------------------------------------- 
 *  Descriptions:  platform豸עᣬmemԴռúͼĴӳ
 */ 
static int s3c2440_i2c_probe(struct device *dev)
{
	struct resource *res;
	int size;
	struct platform_device *pdev = to_platform_device(dev);
	int ret;

	i2c->dev = dev;
	i2c->clk = clk_get(dev, "i2c");
	if (IS_ERR(i2c->clk)) 
	{
		dev_err(dev, "cannot get clock\n");
		return -1;
	}
	clk_use(i2c->clk);
	clk_enable(i2c->clk);

	//ĴռõmemԴ
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL)
	{
		printk(KERN_INFO " get i2c reg memory region failed\n");
		return -ENOENT;
	}
	size = (res->end-res->start)+1;
	i2c->ioarea = request_mem_region(res->start, size, pdev->name);
	if (i2c->ioarea  == NULL)
	{
		printk(KERN_INFO  "failed to get memory region\n");
		return -ENOENT;
	}
	 //ӳi2c Ĵ
	i2c->regs = ioremap(res->start, size);
	if (i2c->regs == 0) 
	{
		printk(KERN_INFO  "failed to ioremap() region\n");
		return -EINVAL;
	}

	if (init_s3c2440_i2c()!= 0)
	{
		return -ENOENT;
	}
	res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (res == NULL)
	{
		printk(dev, "cannot find IRQ\n");
		return  -ENOENT;
	}
	ret = request_irq(res->start, i2c_int_hndlr, SA_INTERRUPT, pdev->name, i2c);
	if (ret != 0) 
	{
		printk(DEVICE_NAME ":cannot request irq for i2c\n");
		return  -ENOENT;
		
	}
	DPRINTK("rIICCON :0X%x\n", readl(rIICCON));
	return 0;
}

static int i2c_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = -1;
	
	if (_IOC_TYPE(cmd) != DEV_IOC_MAGIC)
	{
		return -ENOTTY;
	}
	if (_IOC_NR(cmd) > DEV_IOC_MAXNR) 
	{
		return -ENOTTY;
	}

	switch (cmd)
	{
		case DEV_IOC_SET_I2C_BANDRATE:
		break;
		case DEV_IOC_SET_I2C_SLAVE_ADDR:
			slave_i2c_addr = arg;
			DPRINTK("slave i2c addr:0x%x\n", slave_i2c_addr);
		break;
		default:
			printk(DEVICE_NAME ":error ioctl cmd\n");
			ret = -2;
		break;
	}

	return ret;
}
static int s3c2440_i2c_remove(struct device *dev)
{

	if (i2c->clk != NULL && !IS_ERR(i2c->clk)) 
	{
		clk_disable(i2c->clk);
		clk_unuse(i2c->clk);
		clk_put(i2c->clk);
		i2c->clk = NULL;
	}
	
	if (i2c->regs != NULL) 
	{
		iounmap(i2c->regs);
		i2c->regs = NULL;
	}
	
	if (i2c->ioarea != NULL) 
	{
		release_resource(i2c->ioarea);
		kfree(i2c->ioarea);
		i2c->ioarea = NULL;
	}

	dev_set_drvdata(dev, NULL);
	return 0;
}

#define s3c2440_i2c_resume NULL

static struct device_driver s3c2440_i2c_driver = 
{
	.name		= "s3c2440-i2c",
	.bus		= &platform_bus_type,
	.probe		= s3c2440_i2c_probe,
	.remove		= s3c2440_i2c_remove,
	.resume		= s3c2440_i2c_resume,
};

static struct file_operations i2c_fops = 
{
	owner:	THIS_MODULE,
	ioctl:  i2c_ioctl,
	write:  s3c2440_i2c_write,
	read:   s3c2440_i2c_read,
};

static void write_driver_log(void)
{
	 unsigned char  buf[50];
	 int fd;

	 memset(buf, 0, sizeof(buf));
	 memcpy(buf, DRIVER_VERSION, sizeof(DRIVER_VERSION));
BEGIN_KMEM
	 fd = sys_open(DRIVER_LOG_FILE, O_RDWR | O_APPEND, 0);
	 if(fd>=0)
	 {
		  sys_write(fd, buf, strlen(buf));
		  sys_close(fd);
	 }
	 else
	 	printk("open driver log file error!!\n");
END_KMEM

}


static int __init s3c2440_i2c_adap_init(void)
{
	int ret;
	unsigned char data;
	
 	driver_register(&s3c2440_i2c_driver); 
	ret = register_chrdev(DEVICE_MAJOR, DEVICE_NAME, &i2c_fops);
	if (ret <0)
	{
		printk(DEVICE_NAME ": can't register device\n");
		return ret;
	}

	devfs_mk_cdev(MKDEV(DEVICE_MAJOR, DEVICE_MINOR), S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP, DEVICE_NAME);
	//i2c_write_byte( 0x0, 0x37);
	//i2c_write_byte( 0x1, 0x3d);
	//i2c_read_byte( 0x1, &data);
	write_driver_log();
	printk(DEVICE_NAME ": xx install ok\n");
	
	//i2c_write_byte( 0x0, 0x37);
	return 0;
	
}

static void __exit s3c2440_i2c_adap_exit(void)
{
	free_irq(IRQ_IIC, i2c);
	driver_unregister(&s3c2440_i2c_driver);
	devfs_remove(DEVICE_NAME);
	unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
	printk(DEVICE_NAME ":unregist ok\n");
}

module_init(s3c2440_i2c_adap_init);
module_exit(s3c2440_i2c_adap_exit);

MODULE_LICENSE("GPL");



