/********************************************************
 * FILE info: s3c2440io-bank.c 
 *               2440 gpiobank
 *               ںlinux-2.6.13
 * Date     : 20081104
*********************************************************/


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/arch/map.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/arch/our_s3c2440_regs.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-mem.h>
#include <asm/arch/regs-clock.h>

#define DRIVER_VERSION  "[IOBANK] 1.00.0(090903)\n"
#define DRIVER_LOG_FILE "/dso/driver/driver.log"
#define BEGIN_KMEM {mm_segment_t old_fs = get_fs();set_fs(KERNEL_DS);
#define END_KMEM 	set_fs(old_fs);} 

#define DEBUG 1
#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(fmt, ##args)
#else
#define DPRINTK(fmt, args...)
#endif


#define DEVICE_NAME		"dso-iobank"
#define DEVICE_MAJOR	(247)
#define DEVICE_MINOR	(0)



#define DEV_IOC_MAGIC			'i'
#define DEV_IOC_SET_OUT_L		_IO(DEV_IOC_MAGIC, 0)
#define DEV_IOC_SET_OUT_H		_IO(DEV_IOC_MAGIC, 1)
#define DEV_IOC_SET_IN			_IO(DEV_IOC_MAGIC, 2)
#define DEV_BANK_SET_WIDTH8		_IO(DEV_IOC_MAGIC, 3)
#define DEV_BANK_SET_WIDTH16	_IO(DEV_IOC_MAGIC, 4)
#define DEV_BANK_SET_WIDTH32	_IO(DEV_IOC_MAGIC, 5)




#define GPIO_MAJOR(x)		(((x) & 0x0f00) >> 8)
#define GPIO_MINOR(x)		((x) & 0x1f)
#define OUTPUT_TYPE			(0)
#define INPUT_TYPE			(1)


#define write_0_gpio_bit(x)   do {GPDAT(GRAB_PORT((x))) &= ~(0x1 << GRAB_OFS((x)));}while(0)
#define write_1_gpio_bit(x)   do {GPDAT(GRAB_PORT((x))) |=  (0x1 << GRAB_OFS((x)));}while(0)


enum {
	//BANK0 = 0xf000,
	BANK1 = 0xf001,
	BANK2 = 0xf002,
	BANK3 = 0xf003,
	BANK4 = 0xf004,
	BANK5 = 0xf005,
	BANK6 = 0xf006,
	BANK7 = 0xf007,
};
const unsigned long GPIO_A[] = {
	GPIO_A0,  GPIO_A1,  GPIO_A2,  GPIO_A3,  GPIO_A4,
	GPIO_A5,  GPIO_A6,  GPIO_A7,  GPIO_A8,  GPIO_A9,
	GPIO_A10, GPIO_A11, GPIO_A12, GPIO_A13, GPIO_A14,
	GPIO_A15, GPIO_A16, GPIO_A17, GPIO_A18, GPIO_A19,
	GPIO_A20, GPIO_A21, GPIO_A22, GPIO_A23, GPIO_A24,
};
const unsigned long GPIO_B[] = {
	GPIO_B0,  GPIO_B1,  GPIO_B2,  GPIO_B3,  GPIO_B4,
	GPIO_B5,  GPIO_B6,  GPIO_B7,  GPIO_B8,  GPIO_B9,
	GPIO_B10,
};
const unsigned long GPIO_C[] = {
	GPIO_C0,  GPIO_C1,  GPIO_C2,  GPIO_C3,  GPIO_C4,
	GPIO_C5,  GPIO_C6,  GPIO_C7,  GPIO_C8,  GPIO_C9,
	GPIO_C10, GPIO_C11, GPIO_C12, GPIO_C13, GPIO_C14,
	GPIO_C15,
};
const unsigned long GPIO_D[] = {
	GPIO_D0,  GPIO_D1,  GPIO_D2,  GPIO_D3,  GPIO_D4,
	GPIO_D5,  GPIO_D6,  GPIO_D7,  GPIO_D8,  GPIO_D9,
	GPIO_D10, GPIO_D11, GPIO_D12, GPIO_D13, GPIO_D14,
	GPIO_D15,
};
const unsigned long GPIO_E[] = {
	GPIO_E0,  GPIO_E1,  GPIO_E2,  GPIO_E3,  GPIO_E4,
	GPIO_E5,  GPIO_E6,  GPIO_E7,  GPIO_E8,  GPIO_E9,
	GPIO_E10, GPIO_E11, GPIO_E12, GPIO_E13, GPIO_E14,
	GPIO_E15,
};
const unsigned long GPIO_F[] = {
	GPIO_F0,  GPIO_F1,  GPIO_F2,  GPIO_F3,  GPIO_F4,
	GPIO_F5,  GPIO_F6,  GPIO_F7,
};
const unsigned long GPIO_G[] = {
	GPIO_G0,  GPIO_G1,  GPIO_G2,  GPIO_G3,  GPIO_G4,
	GPIO_G5,  GPIO_G6,  GPIO_G7,  GPIO_G8,  GPIO_G9,
	GPIO_G10, GPIO_G11, GPIO_G12, GPIO_G13, GPIO_G14,
	GPIO_G15,
};
const unsigned long GPIO_H[] = {
	GPIO_H0,  GPIO_H1,  GPIO_H2,  GPIO_H3,  GPIO_H4,
	GPIO_H5,  GPIO_H6,  GPIO_H7,  GPIO_H8,  
};

const unsigned long GPIO_J[] = {	
	GPIO_J0,  GPIO_J1,  GPIO_J2,  GPIO_J3,  GPIO_J4,
	GPIO_J5,  GPIO_J6,  GPIO_J7,  GPIO_J8,  GPIO_J9,
	GPIO_J10, GPIO_J11, GPIO_J12,  
};
	
const unsigned long *gpio_array_pt[] = {
	GPIO_A,
	GPIO_B,
	GPIO_C,
	GPIO_D,
	GPIO_E,
	GPIO_F,
	GPIO_G,
	GPIO_H,
	GPIO_J,
};
const int gpio_counts[] = {
	25, 11, 16, 16, 16, 8, 16, 9, 13, 
};



/*
 * bankݿ
 *
 */
static void set_bank_width(int bank_num, int bank_width)
{
	bank_width = bank_width >> 4;

	writel(readl(S3C2410_BWSCON)&(~(0x3<<(bank_num*4))), S3C2410_BWSCON);
	writel(readl(S3C2410_BWSCON)|(bank_width<<(bank_num*4)), S3C2410_BWSCON);
	DPRINTK("BWSCON:0x%x\n", readl(S3C2410_BWSCON));	
}


static int io_bank_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int flags;
	int bank_num = 0;
	int gpio_name;
	int gpio_major = GPIO_MAJOR(arg);
	int gpio_minor = GPIO_MINOR(arg);
	
	//local_irq_save(flags);
	
	if ((arg < 0xf000) && (arg >= 0) )
	{
		if(gpio_major < sizeof(gpio_counts)/sizeof(gpio_counts[0]))
		{
			if (gpio_minor < gpio_counts[gpio_major])
			{
				gpio_name = gpio_array_pt[gpio_major][gpio_minor];
			}	
		}
		else 
			return -1;
	}
	else if (arg > 0xf000)
	{
		bank_num = arg&0xf;
	}
	else
		return -1;

	switch (cmd)
	{
		case DEV_IOC_SET_OUT_H:
			DPRINTK("gpio_major_n = %d, io_minor_n = %d, output 1\n", gpio_major, gpio_minor);
			set_gpio_ctrl(gpio_name | GPIO_PULLUP_EN | GPIO_MODE_OUT);
			write_1_gpio_bit(gpio_name);
			break;
		case DEV_IOC_SET_OUT_L:
			DPRINTK("gpio_major_n = %d, io_minor_n = %d, output 0\n", gpio_major, gpio_minor);
			set_gpio_ctrl(gpio_name | GPIO_PULLUP_EN | GPIO_MODE_OUT);
			write_0_gpio_bit(gpio_name);
			break;
		case  DEV_IOC_SET_IN:
			DPRINTK("gpio_major_n = %d, io_minor_n = %d. input\n", gpio_major, gpio_minor);
			set_gpio_ctrl(gpio_name | GPIO_MODE_IN);
			break;
		case  DEV_BANK_SET_WIDTH8:
			set_bank_width(bank_num, 8);
			break;
		case  DEV_BANK_SET_WIDTH16:
			set_bank_width(bank_num, 16);
			break;
		case  DEV_BANK_SET_WIDTH32:
			set_bank_width(bank_num, 32);
			break;
		default:
			printk("error cmd\n");
			return -2;
		break;
		
		
		
	}
	
	
	//local_irq_restore(flags);
	return 0;
}



static struct file_operations io_bank_fops =
{
	owner:	THIS_MODULE,
	ioctl:  io_bank_ioctl,
	
};





static void write_driver_log(void)
{
	 unsigned char  buf[100];
	 int fd;
	 unsigned char str[50];

	 memset(buf, 0, sizeof(buf));
	 memset(str, 0, sizeof(str));
	 memcpy(buf, DRIVER_VERSION, sizeof(DRIVER_VERSION));
	  if((readl(S3C2410_MPLLCON)&0xfffff) == 0x44011)
	  {
	  	if(readl(S3C2410_CLKDIVN)&0x07 == 0x7)
			strcpy(str, "CPU 300MHZ BUS 100MHZ\n");
		else
			strcpy(str, "CPU 300MHZ BUS UNKONW\n");
	  }
	  else
	  	strcpy(str, "CPU 400MHZ BUS 100MHZ\n");
	  printk(str);
	  
	strcat(buf, str);
BEGIN_KMEM
	 fd = sys_open(DRIVER_LOG_FILE, O_RDWR | O_APPEND, 0);
	 if(fd>=0)
	 {
		  sys_write(fd, buf, strlen(buf));
		  sys_close(fd);
	 }
	 else
	 	printk("open driver log file error!!\n");
END_KMEM

}

static int __init io_bank_init( void )
{
	int ret;

	ret = register_chrdev(DEVICE_MAJOR, DEVICE_NAME, &io_bank_fops);
	if (ret <0)
	{
		printk(DEVICE_NAME ": can't register device\n");
		return ret;
	}

	devfs_mk_cdev(MKDEV(DEVICE_MAJOR, DEVICE_MINOR), S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP, DEVICE_NAME);
	write_driver_log();
	printk(DEVICE_NAME ": install ok\n");
	return 0;
	
}


static void __exit io_bank_exit( void )
{
	printk(DEVICE_NAME ": uninstall\n");
	devfs_remove(DEVICE_NAME);
	unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
	
}



module_init( io_bank_init );
module_exit( io_bank_exit );
MODULE_LICENSE("GPL");    
MODULE_AUTHOR("TEKWAY");


